// Copyright (c) 2013 Adobe Systems Incorporated. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
import eve from './eve.js'
import { Snap, SnapElement } from './svg.js'
import { is } from './utils.js'
import { doc, win } from './lib/constants.js'

let elproto = SnapElement.prototype,
  has = 'hasOwnProperty',
  supportsTouch =
    'ontouchstart' in window ||
    window.TouchEvent ||
    (window.DocumentTouch && document instanceof DocumentTouch),
  events = [
    'click',
    'dblclick',
    'mousedown',
    'mousemove',
    'mouseout',
    'mouseover',
    'mouseup',
    'touchstart',
    'touchmove',
    'touchend',
    'touchcancel'
  ],
  touchMap = {
    mousedown: 'touchstart',
    mousemove: 'touchmove',
    mouseup: 'touchend'
  },
  getScroll = function (xy, el) {
    let name = xy == 'y' ? 'scrollTop' : 'scrollLeft',
      doc = el && el.node ? el.node.ownerDocument : doc
    return doc[name in doc.documentElement ? 'documentElement' : 'body'][name]
  },
  preventDefault = function () {
    this.returnValue = false
  },
  preventTouch = function () {
    return this.originalEvent.preventDefault()
  },
  stopPropagation = function () {
    this.cancelBubble = true
  },
  stopTouch = function () {
    return this.originalEvent.stopPropagation()
  },
  addEvent = function (obj, type, fn, element) {
    let realName = supportsTouch && touchMap[type] ? touchMap[type] : type,
      f = function (e) {
        let scrollY = getScroll('y', element),
          scrollX = getScroll('x', element)
        if (supportsTouch && touchMap[has](type)) {
          for (
            let i = 0, ii = e.targetTouches && e.targetTouches.length;
            i < ii;
            i++
          ) {
            if (
              e.targetTouches[i].target == obj ||
              obj.contains(e.targetTouches[i].target)
            ) {
              let olde = e
              e = e.targetTouches[i]
              e.originalEvent = olde
              e.preventDefault = preventTouch
              e.stopPropagation = stopTouch
              break
            }
          }
        }
        let x = e.clientX + scrollX,
          y = e.clientY + scrollY
        return fn.call(element, e, x, y)
      }

    if (type !== realName) {
      obj.addEventListener(type, f, false)
    }

    obj.addEventListener(realName, f, false)

    return function () {
      if (type !== realName) {
        obj.removeEventListener(type, f, false)
      }

      obj.removeEventListener(realName, f, false)
      return true
    }
  },
  drag = [],
  dragMove = function (e) {
    let x = e.clientX,
      y = e.clientY,
      scrollY = getScroll('y'),
      scrollX = getScroll('x'),
      dragi,
      j = drag.length
    while (j--) {
      dragi = drag[j]
      if (supportsTouch) {
        let i = e.touches && e.touches.length,
          touch
        while (i--) {
          touch = e.touches[i]
          if (
            touch.identifier == dragi.el._drag.id ||
            dragi.el.node.contains(touch.target)
          ) {
            x = touch.clientX
            y = touch.clientY
            ;(e.originalEvent ? e.originalEvent : e).preventDefault()
            break
          }
        }
      } else {
        e.preventDefault()
      }
      let node = dragi.el.node,
        o,
        next = node.nextSibling,
        parent = node.parentNode,
        display = node.style.display
      // win.opera && parent.removeChild(node);
      // node.style.display = "none";
      // o = dragi.el.paper.getElementByPoint(x, y);
      // node.style.display = display;
      // win.opera && (next ? parent.insertBefore(node, next) : parent.appendChild(node));
      // o && eve("snap.drag.over." + dragi.el.id, dragi.el, o);
      x += scrollX
      y += scrollY
      eve(
        'snap.drag.move.' + dragi.el.id,
        dragi.move_scope || dragi.el,
        x - dragi.el._drag.x,
        y - dragi.el._drag.y,
        x,
        y,
        e
      )
    }
  },
  dragUp = function (e) {
    Snap.unmousemove(dragMove).unmouseup(dragUp)
    let i = drag.length,
      dragi
    while (i--) {
      dragi = drag[i]
      dragi.el._drag = {}
      eve(
        'snap.drag.end.' + dragi.el.id,
        dragi.end_scope || dragi.start_scope || dragi.move_scope || dragi.el,
        e
      )
      eve.off('snap.drag.*.' + dragi.el.id)
    }
    drag = []
  }

for (let i = events.length; i--; ) {
  let eventName = events[i]

  Snap[eventName] = elproto[eventName] = function (fn, scope) {
    if (is(fn, 'function')) {
      this.events = this.events || []
      this.events.push({
        name: eventName,
        f: fn,
        unbind: addEvent(this.node || document, eventName, fn, scope || this)
      })
    } else {
      for (let i = 0, ii = this.events.length; i < ii; i++)
        if (this.events[i].name == eventName) {
          try {
            this.events[i].f.call(this)
          } catch (e) {}
        }
    }
    return this
  }

  Snap['un' + eventName] = elproto['un' + eventName] = function (fn) {
    let events = this.events || [],
      l = events.length
    while (l--)
      if (events[l].name == eventName && (events[l].f == fn || !fn)) {
        events[l].unbind()
        events.splice(l, 1)
        !events.length && delete this.events
        return this
      }
    return this
  }
}
/***
 * Adds hover event handlers to the element
 * f_in (function) handler for hover in
 * f_out (function) handler for hover out
 * icontext (object) #optional context for hover in handler
 * ocontext (object) #optional context for hover out handler
 */
elproto.hover = function (f_in, f_out, scope_in, scope_out) {
  return this.mouseover(f_in, scope_in).mouseout(f_out, scope_out || scope_in)
}
/**
 * Removes hover event handlers from the element
 * f_in (function) handler for hover in
 * f_out (function) handler for hover out
 */
elproto.unhover = function (f_in, f_out) {
  return this.unmouseover(f_in).unmouseout(f_out)
}
let draggable = []

elproto.drag = function (
  onmove,
  onstart,
  onend,
  move_scope,
  start_scope,
  end_scope
) {
  let el = this
  if (!arguments.length) {
    let origTransform
    return el.drag(
      function (dx, dy) {
        this.attr({
          transform: origTransform + (origTransform ? 'T' : 't') + [dx, dy]
        })
      },
      function () {
        origTransform = this.transform().local
      }
    )
  }
  function start(e, x, y) {
    ;(e.originalEvent || e).preventDefault()
    el._drag.x = x
    el._drag.y = y
    el._drag.id = e.identifier
    !drag.length && Snap.mousemove(dragMove).mouseup(dragUp)
    drag.push({
      el: el,
      move_scope: move_scope,
      start_scope: start_scope,
      end_scope: end_scope
    })
    onstart && eve.on('snap.drag.start.' + el.id, onstart)
    onmove && eve.on('snap.drag.move.' + el.id, onmove)
    onend && eve.on('snap.drag.end.' + el.id, onend)
    eve('snap.drag.start.' + el.id, start_scope || move_scope || el, x, y, e)
  }
  function init(e, x, y) {
    eve('snap.draginit.' + el.id, el, e, x, y)
  }
  eve.on('snap.draginit.' + el.id, start)
  el._drag = {}
  draggable.push({ el: el, start: start, init: init })
  el.mousedown(init)
  return el
}

elproto.undrag = function () {
  let i = draggable.length
  while (i--)
    if (draggable[i].el == this) {
      this.unmousedown(draggable[i].init)
      draggable.splice(i, 1)
      eve.unbind('snap.drag.*.' + this.id)
      eve.unbind('snap.draginit.' + this.id)
    }
  !draggable.length && Snap.unmousemove(dragMove).unmouseup(dragUp)
  return this
}
