/**
 * {}
 * @author yutent<yutent.io@gmail.com>
 * @date 2024/03/07 15:22:51
 */

import eve from './eve.js'
import { Snap, SnapElement, Paper, make } from './svg.js'
import { $, is, uuid, preload } from './utils.js'
import { doc, win } from './lib/constants.js'

let proto = Paper.prototype

// gradients
;(function () {
  // gradients' helpers
  /**
   * Only for gradients!
   * Returns array of gradient stops elements.
   */
  function Gstops() {
    return this.selectAll('stop')
  }
  /**
   * Only for gradients!
   * Adds another stop to the gradient.
   * color (string) stops color
   * offset (number) stops offset 0..100
   */
  function GaddStop(color, offset) {
    let stop = $('stop'),
      attr = {
        offset: +offset + '%'
      }
    color = parseColor(color)
    attr['stop-color'] = color.hex
    if (color.opacity < 1) {
      attr['stop-opacity'] = color.opacity
    }
    $(stop, attr)
    let stops = this.stops(),
      inserted
    for (let i = 0; i < stops.length; i++) {
      let stopOffset = parseFloat(stops[i].attr('offset'))
      if (stopOffset > offset) {
        this.node.insertBefore(stop, stops[i].node)
        inserted = true
        break
      }
    }
    if (!inserted) {
      this.node.appendChild(stop)
    }
    return this
  }
  function GgetBBox() {
    if (this.type == 'linearGradient') {
      let x1 = $(this.node, 'x1') || 0,
        x2 = $(this.node, 'x2') || 1,
        y1 = $(this.node, 'y1') || 0,
        y2 = $(this.node, 'y2') || 0
      return Snap._.box(x1, y1, math.abs(x2 - x1), math.abs(y2 - y1))
    } else {
      let cx = this.node.cx || 0.5,
        cy = this.node.cy || 0.5,
        r = this.node.r || 0
      return Snap._.box(cx - r, cy - r, r * 2, r * 2)
    }
  }
  /**
   * Only for gradients!
   * Updates stops of the gradient based on passed gradient descriptor. See @Ppaer.gradient
   - str (string) gradient descriptor part after `()`.
    = (object) gradient element
    | let g = paper.gradient("l(0, 0, 1, 1)#000-#f00-#fff");
    | g.setStops("#fff-#000-#f00-#fc0");
  */
  function GsetStops(str) {
    let grad = str,
      stops = this.stops()
    if (typeof str == 'string') {
      grad = eve(
        'snap.util.grad.parse',
        null,
        'l(0,0,0,1)' + str
      ).firstDefined().stops
    }
    if (!is(grad, 'array')) {
      return
    }
    for (let i = 0; i < stops.length; i++) {
      if (grad[i]) {
        let color = parseColor(grad[i].color),
          attr = { offset: grad[i].offset + '%' }
        attr['stop-color'] = color.hex
        if (color.opacity < 1) {
          attr['stop-opacity'] = color.opacity
        }
        stops[i].attr(attr)
      } else {
        stops[i].remove()
      }
    }
    for (i = stops.length; i < grad.length; i++) {
      this.addStop(grad[i].color, grad[i].offset)
    }
    return this
  }
  function gradient(defs, str) {
    let grad = eve('snap.util.grad.parse', null, str).firstDefined(),
      el
    if (!grad) {
      return null
    }
    grad.params.unshift(defs)
    if (grad.type.toLowerCase() == 'l') {
      el = gradientLinear.apply(0, grad.params)
    } else {
      el = gradientRadial.apply(0, grad.params)
    }
    if (grad.type != grad.type.toLowerCase()) {
      $(el.node, {
        gradientUnits: 'userSpaceOnUse'
      })
    }
    let stops = grad.stops,
      len = stops.length
    for (let i = 0; i < len; i++) {
      let stop = stops[i]
      el.addStop(stop.color, stop.offset)
    }
    return el
  }
  function gradientLinear(defs, x1, y1, x2, y2) {
    let el = make('linearGradient', defs)
    el.stops = Gstops
    el.addStop = GaddStop
    el.getBBox = GgetBBox
    el.setStops = GsetStops
    if (x1 != null) {
      $(el.node, {
        x1: x1,
        y1: y1,
        x2: x2,
        y2: y2
      })
    }
    return el
  }
  function gradientRadial(defs, cx, cy, r, fx, fy) {
    let el = make('radialGradient', defs)
    el.stops = Gstops
    el.addStop = GaddStop
    el.getBBox = GgetBBox
    if (cx != null) {
      $(el.node, {
        cx: cx,
        cy: cy,
        r: r
      })
    }
    if (fx != null && fy != null) {
      $(el.node, {
        fx: fx,
        fy: fy
      })
    }
    return el
  }

  proto.gradient = function (str) {
    return gradient(this.defs, str)
  }
  proto.gradientLinear = function (x1, y1, x2, y2) {
    return gradientLinear(this.defs, x1, y1, x2, y2)
  }
  proto.gradientRadial = function (cx, cy, r, fx, fy) {
    return gradientRadial(this.defs, cx, cy, r, fx, fy)
  }
})()
